#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot._compat import String
from datarobot.errors import ClientError
from datarobot.utils import get_id_from_response
from datarobot.utils.pagination import unpaginate

from ..api_object import APIObject

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models import Job

    class Score(TypedDict):
        label: str
        value: float


DEFAULT_BATCH_SIZE = 100


class ExternalScores(APIObject):
    """Metric scores on prediction dataset with target or actual value column in unsupervised
    case. Contains project metrics for supervised and special classification metrics set for
    unsupervised projects.

    .. versionadded:: v2.21

    Attributes
    ----------
    project_id: str
        id of the project the model belongs to
    model_id: str
        id of the model
    dataset_id: str
        id of the prediction dataset with target or actual value column for unsupervised case
    actual_value_column: Optional[str]
        For unsupervised projects only.
        Actual value column which was used to calculate the classification metrics and
        insights on the prediction dataset.
    scores: list of dicts in a form of {'label': metric_name, 'value': score}
        Scores on the dataset.


    Examples
    --------

    List all scores for a dataset

    .. code-block:: python

        from datarobot.models.external_dataset_scores_insights.external_scores import ExternalScores
        scores = ExternalScores.list(project_id, dataset_id=dataset_id)

    """

    _path = "projects/{project_id}/externalScores/"

    _converter = t.Dict(
        {
            t.Key("project_id"): String(),
            t.Key("model_id"): String(),
            t.Key("dataset_id"): String(),
            t.Key("actual_value_column", optional=True): t.Or(String(), t.Null()),
            t.Key("scores"): t.List(t.Dict({t.Key("label"): String(), t.Key("value"): t.Float()})),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id: str,
        scores: List[Score],
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        actual_value_column: Optional[str] = None,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.actual_value_column = actual_value_column
        self.scores = scores

    @classmethod
    def create(
        cls,
        project_id: str,
        model_id: str,
        dataset_id: str,
        actual_value_column: Optional[str] = None,
    ) -> Job:
        """Compute an external dataset insights for the specified model.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which insights is requested
        dataset_id : str
            id of the dataset for which insights is requested
        actual_value_column : Optional[str]
            actual values column label, for unsupervised projects only


        Returns
        -------
        job : Job
            an instance of created async job
        """

        from datarobot.models import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        payload = {"modelId": model_id, "datasetId": dataset_id}
        if actual_value_column:
            payload["actualValueColumn"] = actual_value_column
        url = cls._path.format(project_id=project_id)
        response = cls._client.post(url, data=payload)
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    @classmethod
    def list(
        cls,
        project_id: str,
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> List[ExternalScores]:
        """Fetch external scores list for the project and optionally for model and dataset.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: Optional[str]
            if specified, only scores for this model will be retrieved
        dataset_id: Optional[str]
            if specified, only scores for this dataset will be retrieved
        offset: Optional[int]
            this many results will be skipped, default: 0
        limit: Optional[int]
            at most this many results are returned, default: 100, max 1000.
            To return all results, specify 0

        Returns
        -------
            A list of :py:class:`External Scores <datarobot.ExternalScores>` objects

        """
        params: Dict[str, Union[str, int]] = {"limit": limit, "offset": offset}
        if model_id:
            params["modelId"] = model_id
        if dataset_id:
            params["datasetId"] = dataset_id

        url = cls._path.format(project_id=project_id)
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            return [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)]
        r_data = cls._client.get(url, params=params).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, project_id: str, model_id: str, dataset_id: str) -> ExternalScores:
        """Retrieve external scores for the project, model and dataset.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: str
            if specified, only scores for this model will be retrieved
        dataset_id: str
            if specified, only scores for this dataset will be retrieved

        Returns
        -------
            :py:class:`External Scores <datarobot.ExternalScores>` object

        """
        if model_id is None:
            raise ValueError("model_id must be specified")
        if dataset_id is None:
            raise ValueError("dataset_id must be specified")
        scores = cls.list(project_id, model_id=model_id, dataset_id=dataset_id)
        if not scores:
            raise ClientError("Requested scores do not exist.", 404)
        return scores[0]

    def __repr__(self) -> str:
        return (
            "Scores(project_id={}, model_id={}, dataset_id={}, "
            "scores={}, actual_value_column={})"
        ).format(
            self.project_id,
            self.model_id,
            self.dataset_id,
            self.scores,
            self.actual_value_column,
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import POST_EDA2_STAGES
from datarobot.models.api_object import APIObject
from datarobot.models.project import Project

__all__ = ["Image", "SampleImage", "DuplicateImage"]

PARAMETER_CAN_NOT_YET_BE_USED_ERROR_MSG = (
    "You are trying to retrieve sample images with certain target values. "
    "The project is not ready for this yet. "
    "Either run this method again after your project is ready for "
    "modeling, or do not pass any target related parameters to this "
    "method."
)

TargetValue = Optional[Union[str, int, float, List[str]]]


class Image(APIObject):
    """An image stored in a project's dataset.

    Attributes
    ----------
    id : str
        Image ID for this image.
    image_type : str
        Image media type. Accessing this may require a server request
        and an associated delay in returning.
    image_bytes : bytes
        Raw bytes of this image. Accessing this may require a server request
        and an associated delay in returning.
    height : int
        Height of the image in pixels.
    width : int
        Width of the image in pixels.
    """

    _get_path = "projects/{project_id}/images/{image_id}/"
    _bytes_path = "projects/{project_id}/images/{image_id}/file/"
    _converter = t.Dict(
        {
            t.Key("image_id"): String(),
            t.Key("height"): Int(),
            t.Key("width"): Int(),
            t.Key("project_id"): String(),
        }
    ).ignore_extra("*")

    def __init__(self, image_id: str, project_id: str, height: int = 0, width: int = 0) -> None:
        self.id = image_id
        self.project_id = project_id
        self.__image_type: Optional[str] = None
        self.__image_bytes: Optional[bytes] = None
        self.height = height
        self.width = width

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.Image("
            "project_id={0.project_id}, "
            "image_id={0.id}, "
            "height={0.height}, "
            "width={0.width})"
        ).format(self)

    @property
    def image_type(self) -> Optional[str]:
        if not self.__image_type:
            self.__get_image_bytes()
        return self.__image_type

    @property
    def image_bytes(self) -> Optional[bytes]:
        if not self.__image_bytes:
            self.__get_image_bytes()
        return self.__image_bytes

    def __get_image_bytes(self) -> None:
        path = self._bytes_path.format(project_id=self.project_id, image_id=self.id)
        r_data = self._client.get(path)
        self.__image_type = r_data.headers.get("Content-Type")
        self.__image_bytes = r_data.content

    @classmethod
    def get(cls, project_id: str, image_id: str) -> "Image":
        """Get a single image object from project.

        Parameters
        ----------
        project_id : str
            Id of the project that contains the images.
        image_id : str
            ID of image to load from the project.
        """
        path = cls._get_path.format(project_id=project_id, image_id=image_id)
        r_data = cls._client.get(path).json()
        r_data["projectId"] = project_id
        ret = cls.from_server_data(r_data)
        return ret


class SampleImage(APIObject):
    """A sample image in a project's dataset.

    If ``Project.stage`` is ``datarobot.enums.PROJECT_STAGE.EDA2`` then
    the ``target_*`` attributes of this class will have values, otherwise
    the values will all be None.

    Attributes
    ----------
    image : Image
        Image object.
    target_value : TargetValue
        Value associated with the ``feature_name``.
    project_id : str
        Id of the project that contains the images.

    """

    _list_pre_eda2_path = "projects/{project_id}/imageSamples/"
    _list_post_eda2_path = "projects/{project_id}/images/"
    _converter = t.Dict(
        {
            t.Key("image_id"): String(),
            t.Key("height"): Int(),
            t.Key("width"): Int(),
            t.Key("target_value", optional=True): t.Or(
                t.String(), t.Int(), t.Float(), t.List(String)
            ),
            t.Key("project_id"): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id: str,
        image_id: str,
        height: int,
        width: int,
        target_value: TargetValue = None,
    ):
        self.image = Image(project_id=project_id, image_id=image_id, height=height, width=width)
        self.project_id = project_id
        self.target_value = target_value

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.SampleImage("
            "project_id={0.project_id}, "
            "image_id={0.image.id}, "
            "target_value={0.target_value})"
        ).format(self)

    @classmethod
    def list(
        cls,
        project_id: str,
        feature_name: str,
        target_value: TargetValue = None,
        target_bin_start: Optional[Union[int, float]] = None,
        target_bin_end: Optional[Union[int, float]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List["SampleImage"]:
        """Get sample images from a project.

        Parameters
        ----------
        project_id : str
            Project that contains the images.
        feature_name : str
            Name of feature column that contains images.
        target_value : TargetValue
            For classification projects - target value to filter images.
            Please note that you can only use this parameter when the project has finished the EDA2
            stage.
        target_bin_start : Optional[Union[int, float]]
            For regression projects - only images corresponding to the target values above
            (inclusive) this value will be returned. Must be specified together with target_bin_end.
            Please note that you can only use this parameter when the project has finished the EDA2
            stage.
        target_bin_end : Optional[Union[int, float]]
            For regression projects - only images corresponding to the target values below
            (exclusive) this value will be returned. Must be specified together with
            target_bin_start.
            Please note that you can only use this parameter when the project has finished the EDA2
            stage.
        offset : Optional[int]
            Number of images to be skipped.
        limit : Optional[int]
            Number of images to be returned.
        """
        project = Project.get(project_id)
        list_params: Dict[str, Any] = {}
        if project.stage in POST_EDA2_STAGES:
            path = cls._list_post_eda2_path.format(project_id=project_id)
            list_params["column"] = feature_name
            if target_value is not None:
                list_params["targetValue"] = target_value
            if target_bin_start is not None:
                list_params["targetBinStart"] = target_bin_start
            if target_bin_end is not None:
                list_params["targetBinEnd"] = target_bin_end
        else:
            path = cls._list_pre_eda2_path.format(project_id=project_id)
            list_params["featureName"] = feature_name
            if (
                target_value is not None
                or target_bin_start is not None
                or target_bin_end is not None
            ):
                # This class uses different routes depending on the project stage. Images with
                # target information are stored as part of the EDA2 workflow. Only the post EDA2
                # route can query for certain targets.
                raise RuntimeError(PARAMETER_CAN_NOT_YET_BE_USED_ERROR_MSG)
        if offset:
            list_params["offset"] = int(offset)
        if limit:
            list_params["limit"] = int(limit)

        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for si_data in r_data["data"]:
            si_data["projectId"] = project_id
            si = cls.from_server_data(si_data)
            ret.append(si)
        return ret


class DuplicateImage(APIObject):
    """An image that was duplicated in the project dataset.

    Attributes
    ----------
    image : Image
        Image object.
    count : int
        Number of times the image was duplicated.
    """

    _list_path = "projects/{project_id}/duplicateImages/{feature_name}/"
    _converter = t.Dict(
        {t.Key("image_id"): String(), t.Key("row_count"): Int(), t.Key("project_id"): String()}
    ).ignore_extra("*")

    def __init__(self, image_id: str, row_count: int, project_id: str):
        self.image = Image(project_id=project_id, image_id=image_id, height=0, width=0)
        self.project_id = project_id
        self.count = row_count

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.DuplicateImage("
            "project_id={0.project_id}, "
            "image_id={0.image.id}, "
            "count={0.count})"
        ).format(self)

    @classmethod
    def list(
        cls,
        project_id: str,
        feature_name: str,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List["DuplicateImage"]:
        """Get all duplicate images in a project.

        Parameters
        ----------
        project_id : str
            Project that contains the images.
        feature_name : str
            Name of feature column that contains images.
        offset : Optional[int]
            Number of images to be skipped.
        limit : Optional[int]
            Number of images to be returned.
        """
        path = cls._list_path.format(project_id=project_id, feature_name=feature_name)
        list_params = {}
        if offset:
            list_params["offset"] = int(offset)
        if limit:
            list_params["limit"] = int(limit)
        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for si_data in r_data["data"]:
            si_data["projectId"] = project_id
            si = cls.from_server_data(si_data)
            ret.append(si)
        return ret



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# pylint: disable=too-many-lines
from __future__ import annotations

from collections import namedtuple
import io
from io import BytesIO, IOBase, StringIO
import os
import tempfile
from typing import (
    Any,
    BinaryIO,
    cast,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

import dateutil
from mypy_extensions import TypedDict
import pandas as pd
import requests
from requests_toolbelt import MultipartEncoder
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import (
    DEFAULT_MAX_WAIT,
    DEFAULT_TIMEOUT,
    FileLocationType,
    LocalSourceType,
    MEGABYTE,
)
from datarobot.errors import InvalidUsageError
from datarobot.mixins.browser_mixin import BrowserMixin
from datarobot.models.api_object import APIObject
from datarobot.models.credential import Credential, CredentialDataSchema
from datarobot.models.feature import DatasetFeature
from datarobot.models.featurelist import DatasetFeaturelist
from datarobot.models.project import Project
from datarobot.models.sharing import SharingAccess
from datarobot.models.use_cases.utils import add_to_use_case, resolve_use_cases, UseCaseLike
from datarobot.utils import assert_single_parameter, dataframe_to_buffer
from datarobot.utils.pagination import unpaginate
from datarobot.utils.source import parse_source_type
from datarobot.utils.sourcedata import list_of_records_to_buffer
from datarobot.utils.waiters import wait_for_async_resolution

if TYPE_CHECKING:
    from datarobot.models.recipe import Recipe


ProjectLocation = namedtuple("ProjectLocation", ["url", "id"])

FeatureTypeCount = namedtuple("FeatureTypeCount", ["count", "feature_type"])


class MaterializationDestination(TypedDict):
    catalog: str
    schema: str
    # Table name to create and materialize the recipe to. This table should not already exist.
    table: str


TDataset = TypeVar("TDataset", bound="Dataset")
TDatasetDetails = TypeVar("TDatasetDetails", bound="DatasetDetails")

_base_dataset_schema = t.Dict(
    {
        t.Key("dataset_id"): String,
        t.Key("version_id"): String,
        t.Key("name"): String,
        t.Key("categories"): t.List(String),
        t.Key("creation_date") >> "created_at": t.Call(dateutil.parser.parse),
        t.Key("created_by", optional=True): t.Or(String, t.Null),
        t.Key("data_persisted", optional=True): t.Bool,
        t.Key("is_data_engine_eligible"): t.Bool,
        t.Key("is_latest_version"): t.Bool,
        t.Key("is_snapshot"): t.Bool,
        t.Key("dataset_size", optional=True) >> "size": Int,
        t.Key("row_count", optional=True): Int,
        t.Key("sample_size", optional=True): t.Or(
            t.Dict({t.Key("type"): String, t.Key("value"): Int}), t.Null
        ),
        t.Key("processing_state"): String,
    }
)


class Dataset(APIObject, BrowserMixin):
    """Represents a Dataset returned from the api/v2/datasets/ endpoints.

    Attributes
    ----------
    id: string
        The ID of this dataset
    name: string
        The name of this dataset in the catalog
    is_latest_version: bool
        Whether this dataset version is the latest version
        of this dataset
    version_id: string
        The object ID of the catalog_version the dataset belongs to
    categories: list(string)
        An array of strings describing the intended use of the dataset. The
        supported options are "TRAINING" and "PREDICTION".
    created_at: string
        The date when the dataset was created
    created_by: string, optional
        Username of the user who created the dataset
    is_snapshot: bool
        Whether the dataset version is an immutable snapshot of data
        which has previously been retrieved and saved to Data_robot
    data_persisted: Optional[bool]
        If true, user is allowed to view extended data profile
        (which includes data statistics like min/max/median/mean, histogram, etc.) and download
        data. If false, download is not allowed and only the data schema (feature names and types)
        will be available.
    is_data_engine_eligible: bool
        Whether this dataset can be
        a data source of a data engine query.
    processing_state: string
        Current ingestion process state of
        the dataset
    row_count: Optional[int]
        The number of rows in the dataset.
    size: Optional[int]
        The size of the dataset as a CSV in bytes.
    sample_size: dict, optional
        The size of data fetched during dataset registration. For example, to fetch the first 95
        rows,  the sample_size value is {'type': 'rows', 'value': 95}.
        Currently only 'rows' type is supported.
    """

    _converter = _base_dataset_schema.allow_extra("*")
    _path = "datasets/"

    def __init__(
        self,
        dataset_id: str,
        version_id: str,
        name: str,
        categories: List[str],
        created_at: str,
        is_data_engine_eligible: bool,
        is_latest_version: bool,
        is_snapshot: bool,
        processing_state: str,
        created_by: Optional[str] = None,
        data_persisted: Optional[bool] = None,
        size: Optional[int] = None,
        row_count: Optional[int] = None,
        recipe_id: Optional[str] = None,
        sample_size: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.id = dataset_id
        self.version_id = version_id
        self.name = name
        self.data_persisted = data_persisted
        self.categories = categories[:]
        self.created_at = created_at
        self.created_by = created_by
        self.is_data_engine_eligible = is_data_engine_eligible
        self.is_latest_version = is_latest_version
        self.is_snapshot = is_snapshot
        self.size = size
        self.row_count = row_count
        self.processing_state = processing_state
        self.recipe_id = recipe_id
        self.sample_size = sample_size

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, id={self.id!r})"

    def get_uri(self) -> str:
        """
        Returns
        -------
        url : str
            Permanent static hyperlink to this dataset in AI Catalog.
        """
        return f"{self._client.domain}/ai-catalog/{self.id}"

    @classmethod
    def upload(  # type: ignore[return]
        cls: Type[TDataset],
        source: Union[str, pd.DataFrame, IOBase],
    ) -> TDataset:
        """This method covers Dataset creation from local materials (file & DataFrame) and a URL.

        Parameters
        ----------
        source: str, pd.DataFrame or file object
            Pass a URL, filepath, file or DataFrame to create and return a Dataset.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data source.

        Raises
        ------
        InvalidUsageError
            If the source parameter cannot be determined to be a URL, filepath, file or DataFrame.

        Examples
        --------
        .. code-block:: python

            # Upload a local file
            dataset_one = Dataset.upload("./data/examples.csv")

            # Create a dataset via URL
            dataset_two = Dataset.upload(
                "https://raw.githubusercontent.com/curran/data/gh-pages/dbpedia/cities/data.csv"
            )

            # Create dataset with a pandas Dataframe
            dataset_three = Dataset.upload(my_df)

            # Create dataset using a local file
            with open("./data/examples.csv", "rb") as file_pointer:
                dataset_four = Dataset.create_from_file(filelike=file_pointer)
        """
        source_type = parse_source_type(source)
        if source_type == FileLocationType.URL:
            return cls.create_from_url(url=cast(str, source))
        elif source_type == FileLocationType.PATH:
            return cls.create_from_file(file_path=cast(str, source))
        elif source_type == LocalSourceType.DATA_FRAME:
            return cls.create_from_in_memory_data(data_frame=source)
        elif source_type == LocalSourceType.FILELIKE:
            return cls.create_from_file(filelike=cast(IOBase, source))

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_file(
        cls: Type[TDataset],
        file_path: Optional[str] = None,
        filelike: Optional[IOBase] = None,
        categories: Optional[List[str]] = None,
        read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from a file. Returns when the dataset has
        been successfully uploaded and processed.

        Warning: This function does not clean up it's open files. If you pass a filelike, you are
        responsible for closing it. If you pass a file_path, this will create a file object from
        the file_path but will not close it.

        Parameters
        ----------
        file_path: string, optional
            The path to the file. This will create a file object pointing to that file but will
            not close it.
        filelike: file, optional
            An open and readable file object.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: Optional[int]
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case ids or a single Use Case id to add this new Dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            A fully armed and operational Dataset
        """
        assert_single_parameter(("filelike", "file_path"), file_path, filelike)

        upload_url = f"{cls._path}fromFile/"
        default_fname = "data.csv"
        if file_path:
            fname = os.path.basename(file_path)
            response = cls._client.build_request_with_file(
                fname=fname,
                file_path=file_path,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )
        else:
            fname = getattr(filelike, "name", default_fname)
            response = cls._client.build_request_with_file(
                fname=fname,
                filelike=filelike,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        dataset = cls.from_location(new_dataset_location)
        if categories:
            dataset.modify(categories=categories)
        return dataset

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_in_memory_data(
        cls: Type[TDataset],
        data_frame: Optional[pd.DataFrame] = None,
        records: Optional[List[Dict[str, Any]]] = None,
        categories: Optional[List[str]] = None,
        read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
        max_wait: int = DEFAULT_MAX_WAIT,
        fname: Optional[
            str
        ] = None,  # This line provided under MIT license copyright (c) 2021 AGEAS SA/NV
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from in-memory data. Returns when the dataset has
        been successfully uploaded and processed.

        The data can be either a pandas DataFrame or a list of dictionaries with identical keys.

        Parameters
        ----------
        data_frame: DataFrame, optional
            The data frame to upload
        records: list[dict], optional
            A list of dictionaries with identical keys to upload
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: Optional[int]
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful
        fname: string, optional
            The file name, "data.csv" by default
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data.

        Raises
        ------
        InvalidUsageError
            If neither a DataFrame or list of records is passed.
        """
        assert_single_parameter(("data_frame", "records"), data_frame, records)
        if data_frame is not None:
            buff = dataframe_to_buffer(data_frame)
        elif records:
            buff = list_of_records_to_buffer(records)
        else:
            raise InvalidUsageError("Must pass either a DataFrame or list or records")
        if fname is not None:  # This line provided under MIT license copyright (c) 2021 AGEAS SA/NV
            buff.name = fname  # This line provided under MIT license copyright (c) 2021 AGEAS SA/NV
        return cls.create_from_file(
            filelike=buff,
            categories=categories,
            read_timeout=read_timeout,
            max_wait=max_wait,
        )

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_url(
        cls: Type[TDataset],
        url: str,
        do_snapshot: Optional[bool] = None,
        persist_data_after_ingestion: Optional[bool] = None,
        categories: Optional[List[str]] = None,
        sample_size: Optional[Dict[str, Any]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored at a url.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        url: string
            The URL to use as the source of data for the dataset being created.
        do_snapshot: Optional[bool]
            If unset, uses the server default: True.
            If true, creates a snapshot dataset; if
            false, creates a remote dataset. Creating snapshots from non-file sources may be
            disabled by the permission, `Disable AI Catalog Snapshots`.
        persist_data_after_ingestion: Optional[bool]
            If unset, uses the server default: True.
            If true, will enforce saving all data
            (for download and sampling) and will allow a user to view extended data profile
            (which includes data statistics like min/max/median/mean, histogram, etc.). If false,
            will not enforce saving data. The data schema (feature names and types) still will be
            available. Specifying this parameter to false and `doSnapshot` to true will result in
            an error.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        sample_size: dict, optional
            The size of data fetched during dataset registration.
            For example, to fetch the first 95 rows,  the sample_size value would be:
            {'type': 'rows', 'value': 95}.
            Currently only 'rows' type is supported.
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful.
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "url": url,
            "do_snapshot": do_snapshot,
            "persist_data_after_ingestion": persist_data_after_ingestion,
            "categories": categories,
            "sample_size": sample_size,
        }
        data = _remove_empty_params(base_data)
        upload_url = f"{cls._path}fromURL/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_project(
        cls: Type[TDataset],
        project_id: str,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new dataset from project data.
        Returns when the dataset has been successfully created.

        Parameters
        ----------
        project_id: string
            The project to create the dataset from.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful.
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            The dataset created from the project dataset.
        """
        base_data = {"project_id": project_id, "categories": categories}
        data = _remove_empty_params(base_data)
        url = f"{cls._path}fromProject/"
        response = cls._client.post(url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_datastage(
        cls: Type[TDataset],
        datastage_id: str,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored as a DataStage.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        datastage_id: string
            The ID of the DataStage to use as the source of data for the dataset being created.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "stage_id": datastage_id,
            "categories": categories,
        }
        data = _remove_empty_params(base_data)
        upload_url = f"{cls._path}fromStage/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_data_source(
        cls: Type[TDataset],
        data_source_id: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        do_snapshot: Optional[bool] = None,
        persist_data_after_ingestion: Optional[bool] = None,
        categories: Optional[List[str]] = None,
        credential_id: Optional[str] = None,
        use_kerberos: Optional[bool] = None,
        credential_data: Optional[Dict[str, str]] = None,
        sample_size: Optional[Dict[str, Any]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored at a DataSource.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.22

        Parameters
        ----------
        data_source_id: string
            The ID of the DataSource to use as the source of data.
        username: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored.
        do_snapshot: Optional[bool]
            If unset, uses the server default: True.
            If true, creates a snapshot dataset; if
            false, creates a remote dataset. Creating snapshots from non-file sources requires may
            be disabled by the permission, `Disable AI Catalog Snapshots`.
        persist_data_after_ingestion: Optional[bool]
            If unset, uses the server default: True.
            If true, will enforce saving all data
            (for download and sampling) and will allow a user to view extended data profile
            (which includes data statistics like min/max/median/mean, histogram, etc.). If false,
            will not enforce saving data. The data schema (feature names and types) still will be
            available. Specifying this parameter to false and `doSnapshot` to true will result in
            an error.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        credential_id: string, optional
            The ID of the set of credentials to
            use instead of user and password. Note that with this change, username and password
            will become optional.
        use_kerberos: Optional[bool]
            If unset, uses the server default: False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.
        sample_size: dict, optional
            The size of data fetched during dataset registration.
            For example, to fetch the first 95 rows,  the sample_size value would be:
            {'type': 'rows', 'value': 95}.
            Currently only 'rows' type is supported.
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful.
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "data_source_id": data_source_id,
            "user": username,
            "password": password,
            "do_snapshot": do_snapshot,
            "persist_data_after_ingestion": persist_data_after_ingestion,
            "categories": categories,
            "credential_id": credential_id,
            "use_kerberos": use_kerberos,
            "credential_data": credential_data,
            "sample_size": sample_size,
        }
        data = _remove_empty_params(base_data)

        if "credential_data" in data:
            data["credential_data"] = CredentialDataSchema(data["credential_data"])

        upload_url = f"{cls._path}fromDataSource/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_query_generator(
        cls: Type[TDataset],
        generator_id: str,
        dataset_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from the query generator.
        Returns when the dataset has been successfully processed. If optional
        parameters are not specified the query is applied to the dataset_id
        and dataset_version_id stored in the query generator. If specified they
        will override the stored dataset_id/dataset_version_id, e.g. to prep a
        prediction dataset.

        Parameters
        ----------
        generator_id: str
            The id of the query generator to use.
        dataset_id: Optional[str]
            The id of the dataset to apply the query to.
        dataset_version_id: Optional[str]
            The id of the dataset version to apply the query to. If not specified the
            latest version associated with dataset_id (if specified) is used.
        max_wait : int
            optional, the maximum number of seconds to wait before giving up.
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case IDs or a single Use Case ID to add this new dataset to. Must be a kwarg.

        Returns
        -------
        response: Dataset
            The Dataset created from the query generator
        """
        url = "dataEngineWorkspaceStates/fromDataEngineQueryGenerator/"
        base_data = {
            "query_generator_id": generator_id,
            "dataset_id": dataset_id,
            "dataset_version_id": dataset_version_id,
        }
        data = _remove_empty_params(base_data)
        response = cls._client.post(url, data=data)
        workspace_id = response.json()["workspaceStateId"]

        url = f"{cls._path}fromDataEngineWorkspaceState/"
        response = cls._client.post(url, data={"workspace_state_id": workspace_id})
        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    @add_to_use_case(allow_multiple=True)
    def create_from_recipe(
        cls: Type[TDataset],
        recipe: Recipe,
        name: Optional[str] = None,
        do_snapshot: Optional[bool] = None,
        persist_data_after_ingestion: Optional[bool] = None,
        categories: Optional[List[str]] = None,
        credential: Optional[Credential] = None,
        use_kerberos: Optional[bool] = None,
        materialization_destination: Optional[MaterializationDestination] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from the recipe.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: 3.6

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "recipe_id": recipe.id,
            "name": name,
            "do_snapshot": do_snapshot,
            "persist_data_after_ingestion": persist_data_after_ingestion,
            "categories": categories,
            "use_kerberos": use_kerberos,
            "materialization_destination": materialization_destination,
        }
        data = _remove_empty_params(base_data)
        if credential is not None:
            data["credential_id"] = credential.credential_id

        upload_url = f"{cls._path}fromRecipe/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def get(cls: Type[TDataset], dataset_id: str) -> TDataset:
        """Get information about a dataset.

        Parameters
        ----------
        dataset_id : string
            the id of the dataset

        Returns
        -------
        dataset : Dataset
            the queried dataset
        """

        path = f"{cls._path}{dataset_id}/"
        return cls.from_location(path)

    @classmethod
    def delete(cls, dataset_id: str) -> None:
        """
        Soft deletes a dataset.  You cannot get it or list it or do actions with it, except for
        un-deleting it.

        Parameters
        ----------
        dataset_id: string
            The id of the dataset to mark for deletion

        Returns
        -------
        None
        """
        path = f"{cls._path}{dataset_id}/"
        cls._client.delete(path)

    @classmethod
    def un_delete(cls, dataset_id: str) -> None:
        """
        Un-deletes a previously deleted dataset.  If the dataset was not deleted, nothing happens.

        Parameters
        ----------
        dataset_id: string
            The id of the dataset to un-delete

        Returns
        -------
        None
        """
        path = f"{cls._path}{dataset_id}/deleted/"
        cls._client.patch(path)

    @classmethod
    def list(
        cls: Type[TDataset],
        category: Optional[str] = None,
        filter_failed: Optional[bool] = None,
        order_by: Optional[str] = None,
        use_cases: Optional[UseCaseLike] = None,
    ) -> List[TDataset]:
        """
        List all datasets a user can view.


        Parameters
        ----------
        category: string, optional
            Optional. If specified, only dataset versions that have the specified category will be
            included in the results. Categories identify the intended use of the dataset; supported
            categories are "TRAINING" and "PREDICTION".

        filter_failed: Optional[bool]
            If unset, uses the server default: False.
            Whether datasets that failed during import should be excluded from the results.
            If True invalid datasets will be excluded.

        order_by: string, optional
            If unset, uses the server default: "-created".
            Sorting order which will be applied to catalog list, valid options are:
            - "created" -- ascending order by creation datetime;
            - "-created" -- descending order by creation datetime.

        use_cases: Union[UseCase, List[UseCase], str, List[str]], optional
            Filter available datasets by a specific Use Case or Cases. Accepts either the entity or the ID.
            If set to [None], the method filters the project's datasets by those not linked to a UseCase.

        Returns
        -------
        list[Dataset]
            a list of datasets the user can view

        """
        return list(
            cls.iterate(
                category=category,
                order_by=order_by,
                filter_failed=filter_failed,
                use_cases=use_cases,
            )
        )

    @classmethod
    def iterate(
        cls: Type[TDataset],
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        category: Optional[str] = None,
        order_by: Optional[str] = None,
        filter_failed: Optional[bool] = None,
        use_cases: Optional[UseCaseLike] = None,
    ) -> Generator[TDataset, None, None]:
        """
        Get an iterator for the requested datasets a user can view.
        This lazily retrieves results. It does not get the next page from the server until the
        current page is exhausted.

        Parameters
        ----------
        offset: Optional[int]
            If set, this many results will be skipped

        limit: Optional[int]
            Specifies the size of each page retrieved from the server.  If unset, uses the server
            default.

        category: string, optional
            Optional. If specified, only dataset versions that have the specified category will be
            included in the results. Categories identify the intended use of the dataset; supported
            categories are "TRAINING" and "PREDICTION".

        filter_failed: Optional[bool]
            If unset, uses the server default: False.
            Whether datasets that failed during import should be excluded from the results.
            If True invalid datasets will be excluded.

        order_by: string, optional
            If unset, uses the server default: "-created".
            Sorting order which will be applied to catalog list, valid options are:
            - "created" -- ascending order by creation datetime;
            - "-created" -- descending order by creation datetime.

        use_cases: Union[UseCase, List[UseCase], str, List[str]], optional
            Filter available datasets by a specific Use Case or Cases. Accepts either the entity or the ID.
            If set to [None], the method filters the project's datasets by those not linked to a UseCase.

        Yields
        ------
        Dataset
            An iterator of the datasets the user can view.

        """
        all_params = {
            "offset": offset,
            "limit": limit,
            "category": category,
            "order_by": order_by,
            "filter_failed": filter_failed,
        }
        all_params = resolve_use_cases(use_cases=use_cases, params=all_params)
        params = _remove_empty_params(all_params)
        _update_filter_failed(params)

        for dataset_json in unpaginate(cls._path, params, cls._client):
            yield cls.from_server_data(dataset_json)

    def update(self) -> None:
        """
        Updates the Dataset attributes in place with the latest information from the server.

        Returns
        -------
        None
        """
        new_dataset = self.get(self.id)
        update_attrs = (
            "name",
            "created_by",
            "created_at",
            "version_id",
            "is_latest_version",
            "is_snapshot",
            "data_persisted",
            "categories",
            "size",
            "row_count",
            "processing_state",
        )
        for attr in update_attrs:
            setattr(self, attr, getattr(new_dataset, attr))

    def modify(
        self,
        name: Optional[str] = None,
        categories: Optional[List[str]] = None,
    ) -> None:
        """
        Modifies the Dataset name and/or categories.  Updates the object in place.

        Parameters
        ----------
        name: string, optional
            The new name of the dataset

        categories: list[string], optional
            A list of strings describing the intended use of the
            dataset. The supported options are "TRAINING" and "PREDICTION". If any
            categories were previously specified for the dataset, they will be overwritten.
            If omitted or None, keep previous categories. To clear them specify []

        Returns
        -------
        None

        """
        if name is None and categories is None:
            return
        # if we omit categories is will clear them
        # so in case where it is None, we want to preserve
        # categories and only clear them when it is []
        if categories is None:
            categories = self.categories

        url = f"{self._path}{self.id}/"
        params = {"name": name, "categories": categories}
        params = _remove_empty_params(params)

        response = self._client.patch(url, data=params)
        data = response.json()
        self.name = data["name"]
        self.categories = data["categories"]

    def share(
        self,
        access_list: List[SharingAccess],
        apply_grant_to_linked_objects: bool = False,
    ) -> None:
        """Modify the ability of users to access this dataset

        Parameters
        ----------
        access_list: list of :class:`SharingAccess <datarobot.SharingAccess>`
            The modifications to make.

        apply_grant_to_linked_objects: bool
            If true for any users being granted access to the dataset, grant the user read access to
            any linked objects such as DataSources and DataStores that may be used by this dataset.
            Ignored if no such objects are relevant for dataset, defaults to False.

        Returns
        -------
        None

        Raises
        ------
        datarobot.ClientError:
            If you do not have permission to share this dataset, if the user you're sharing with
            doesn't exist, if the same user appears multiple times in the access_list, or if these
            changes would leave the dataset without an owner.

        Examples
        --------
        Transfer access to the dataset from old_user@datarobot.com to new_user@datarobot.com

        .. code-block:: python

            from datarobot.enums import SHARING_ROLE
            from datarobot.models.dataset import Dataset
            from datarobot.models.sharing import SharingAccess

            new_access = SharingAccess(
                "new_user@datarobot.com",
                SHARING_ROLE.OWNER,
                can_share=True,
            )
            access_list = [
                SharingAccess(
                    "old_user@datarobot.com",
                    SHARING_ROLE.OWNER,
                    can_share=True,
                    can_use_data=True,
                ),
                new_access,
            ]

            Dataset.get('my-dataset-id').share(access_list)
        """
        payload = {
            "applyGrantToLinkedObjects": apply_grant_to_linked_objects,
            "data": [access.collect_payload() for access in access_list],
        }
        self._client.patch(
            f"{self._path}{self.id}/accessControl/", data=payload, keep_attrs={"role"}
        )

    def get_details(self) -> DatasetDetails:
        """
        Gets the details for this Dataset

        Returns
        -------
        DatasetDetails
        """
        return DatasetDetails.get(self.id)

    def get_all_features(self, order_by: Optional[str] = None) -> List[DatasetFeature]:
        """
        Get a list of all the features for this dataset.

        Parameters
        ----------
        order_by: string, optional
            If unset, uses the server default: 'name'.
            How the features should be ordered. Can be 'name' or 'featureType'.

        Returns
        -------
        list[DatasetFeature]
        """
        return list(self.iterate_all_features(order_by=order_by))

    def iterate_all_features(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> Generator[DatasetFeature, None, None]:
        """
        Get an iterator for the requested features of a dataset.
        This lazily retrieves results. It does not get the next page from the server until the
        current page is exhausted.

        Parameters
        ----------
        offset: Optional[int]
            If set, this many results will be skipped.

        limit: Optional[int]
            Specifies the size of each page retrieved from the server.  If unset, uses the server
            default.

        order_by: string, optional
            If unset, uses the server default: 'name'.
            How the features should be ordered. Can be 'name' or 'featureType'.

        Yields
        ------
        DatasetFeature
        """
        all_params = {
            "offset": offset,
            "limit": limit,
            "order_by": order_by,
        }
        params = _remove_empty_params(all_params)

        url = f"{self._path}{self.id}/allFeaturesDetails/"
        for dataset_json in unpaginate(url, params, self._client):
            yield DatasetFeature.from_server_data(dataset_json)

    def get_featurelists(self) -> List[DatasetFeaturelist]:
        """
        Get DatasetFeaturelists created on this Dataset

        Returns
        -------
        feature_lists: list[DatasetFeaturelist]
        """
        url = f"{self._path}{self.id}/featurelists/"
        params: Dict[str, str] = {}
        result = unpaginate(url, params, self._client)
        return [DatasetFeaturelist.from_server_data(el) for el in result]

    def create_featurelist(self, name: str, features: List[str]) -> DatasetFeaturelist:
        """Create a new dataset featurelist

        Parameters
        ----------
        name : str
            the name of the modeling featurelist to create. Names must be unique within the
            dataset, or the server will return an error.
        features : List[str]
            the names of the features to include in the dataset featurelist. Each feature must
            be a dataset feature.

        Returns
        -------
        featurelist : DatasetFeaturelist
            the newly created featurelist

        Examples
        --------
        .. code-block:: python

            dataset = Dataset.get('1234deadbeeffeeddead4321')
            dataset_features = dataset.get_all_features()
            selected_features = [feat.name for feat in dataset_features][:5]  # select first five
            new_flist = dataset.create_featurelist('Simple Features', selected_features)
        """
        url = f"{self._path}{self.id}/featurelists/"

        payload = {"name": name, "features": features}
        response = self._client.post(url, data=payload)
        return DatasetFeaturelist.from_server_data(response.json())

    def get_file(self, file_path: Optional[str] = None, filelike: Optional[IOBase] = None) -> None:
        """
        Retrieves all the originally uploaded data in CSV form.
        Writes it to either the file or a filelike object that can write bytes.

        Only one of file_path or filelike can be provided and it must be provided as a
        keyword argument (i.e. file_path='path-to-write-to'). If a file-like object is
        provided, the user is responsible for closing it when they are done.

        The user must also have permission to download data.

        Parameters
        ----------
        file_path: string, optional
            The destination to write the file to.
        filelike: file, optional
            A file-like object to write to.  The object must be able to write bytes. The user is
            responsible for closing the object

        Returns
        -------
        None
        """
        assert_single_parameter(("filelike", "file_path"), filelike, file_path)

        response = self._client.get(f"{self._path}{self.id}/file/")
        if file_path:
            with open(file_path, "wb") as f:
                f.write(response.content)
        if filelike:
            filelike.write(response.content)

    def get_as_dataframe(self, low_memory: Optional[bool] = False) -> pd.DataFrame:
        """
        Retrieves all the originally uploaded data in a pandas DataFrame.

        .. versionadded:: v3.0

        Parameters
        ----------
        low_memory: Optional[bool]
            If True, use local files to reduce memory usage which will be slower.

        Returns
        -------
        pd.DataFrame
        """
        if low_memory:
            with tempfile.NamedTemporaryFile(suffix=".csv") as csv_file:
                iter = self._client.get(f"{self._path}{self.id}/file/", stream=True)
                with open(csv_file.name, "wb") as out:
                    for chunk in iter.iter_content(1000):
                        out.write(chunk)
                try:
                    return pd.read_csv(csv_file.name)
                except Exception:
                    # different versions of python and pandas
                    # raise different exceptions
                    with open(csv_file.name, "rb") as inp:
                        dat = inp.read(4)
                    if dat == b"PAR1":
                        return pd.read_parquet(csv_file.name)
                    raise
        else:
            raw_bytes = BytesIO()
            self.get_file(filelike=raw_bytes)
            if raw_bytes.getvalue().startswith(b"PAR1"):
                return pd.read_parquet(raw_bytes)
            data = StringIO(raw_bytes.getvalue().decode())
            return pd.read_csv(data)

    def get_projects(self) -> List[ProjectLocation]:
        """
        Retrieves the Dataset's projects as ProjectLocation named tuples.

        Returns
        -------
        locations: list[ProjectLocation]
        """
        url = f"{self._path}{self.id}/projects/"
        return [ProjectLocation(**kwargs) for kwargs in unpaginate(url, None, self._client)]

    @add_to_use_case(allow_multiple=True)
    def create_project(
        self,
        project_name: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        credential_id: Optional[str] = None,
        use_kerberos: Optional[bool] = None,
        credential_data: Optional[Dict[str, str]] = None,
    ) -> Project:
        """
        Create a :class:`datarobot.models.Project` from this dataset

        Parameters
        ----------
        project_name: string, optional
            The name of the project to be created.
            If not specified, will be "Untitled Project" for database connections, otherwise
            the project name will be based on the file used.
        user: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored
        credential_id: string, optional
            The ID of the set of credentials to use instead of user and password.
        use_kerberos: Optional[bool]
            Server default is False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.
        use_cases: list[UseCase] | UseCase | list[string] | string, optional
            A list of UseCase objects, UseCase object,
            list of Use Case ids or a single Use Case id to add this new Dataset to. Must be a kwarg.

        Returns
        -------
        Project
        """
        return Project.create_from_dataset(
            self.id,
            dataset_version_id=self.version_id,
            project_name=project_name,
            user=user,
            password=password,
            credential_id=credential_id,
            use_kerberos=use_kerberos,
            credential_data=credential_data,
        )

    @classmethod
    def create_version_from_file(
        cls: Type[TDataset],
        dataset_id: str,
        file_path: Optional[str] = None,
        filelike: Optional[IOBase] = None,
        categories: Optional[List[str]] = None,
        read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset version from a file. Returns when the new dataset
        version has been successfully uploaded and processed.

        Warning: This function does not clean up it's open files. If you pass a filelike, you are
        responsible for closing it. If you pass a file_path, this will create a file object from
        the file_path but will not close it.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        file_path: string, optional
            The path to the file. This will create a file object pointing to that file but will
            not close it.
        filelike: file, optional
            An open and readable file object.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: Optional[int]
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            A fully armed and operational Dataset version
        """
        assert_single_parameter(("filelike", "file_path"), file_path, filelike)

        upload_url = f"{cls._path}{dataset_id}/versions/fromFile/"
        default_fname = "data.csv"
        if file_path:
            fname = os.path.basename(file_path)
            response = cls._client.build_request_with_file(
                fname=fname,
                file_path=file_path,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )
        else:
            fname = getattr(filelike, "name", default_fname)
            response = cls._client.build_request_with_file(
                fname=fname,
                filelike=filelike,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        dataset = cls.from_location(new_dataset_location)
        if categories:
            dataset.modify(categories=categories)
        return dataset

    @classmethod
    def create_version_from_in_memory_data(
        cls: Type[TDataset],
        dataset_id: str,
        data_frame: Optional[pd.DataFrame] = None,
        records: Optional[List[Dict[str, Any]]] = None,
        categories: Optional[List[str]] = None,
        read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset version for a dataset from in-memory data.
        Returns when the dataset has been successfully uploaded and processed.

        The data can be either a pandas DataFrame or a list of dictionaries with identical keys.

         .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        data_frame: DataFrame, optional
            The data frame to upload
        records: list[dict], optional
            A list of dictionaries with identical keys to upload
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: Optional[int]
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data

        Raises
        ------
        InvalidUsageError
            If neither a DataFrame or list of records is passed.
        """
        assert_single_parameter(("data_frame", "records"), data_frame, records)
        if data_frame is not None:
            buff = dataframe_to_buffer(data_frame)
        elif records:
            buff = list_of_records_to_buffer(records)
        else:
            raise InvalidUsageError("Must pass either a DataFrame or list or records")
        return cls.create_version_from_file(
            dataset_id,
            filelike=buff,
            categories=categories,
            read_timeout=read_timeout,
            max_wait=max_wait,
        )

    @classmethod
    def create_version_from_url(
        cls: Type[TDataset],
        dataset_id: str,
        url: str,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored at a url for a given dataset.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        url: string
            The URL to use as the source of data for the dataset being created.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        base_data = {
            "url": url,
            "categories": categories,
        }
        data = _remove_empty_params(base_data)
        upload_url = f"{cls._path}{dataset_id}/versions/fromURL/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def create_version_from_datastage(
        cls: Type[TDataset],
        dataset_id: str,
        datastage_id: str,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored as a DataStage for a given dataset.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        datastage_id: string
            The ID of the DataStage to use as the source of data for the dataset being created.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        base_data = {
            "stage_id": datastage_id,
            "categories": categories,
        }
        data = _remove_empty_params(base_data)
        upload_url = f"{cls._path}{dataset_id}/versions/fromStage/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def create_version_from_data_source(
        cls: Type[TDataset],
        dataset_id: str,
        data_source_id: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        categories: Optional[List[str]] = None,
        credential_id: Optional[str] = None,
        use_kerberos: Optional[bool] = None,
        credential_data: Optional[Dict[str, str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> TDataset:
        """
        A blocking call that creates a new Dataset from data stored at a DataSource.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        data_source_id: string
            The ID of the DataSource to use as the source of data.
        username: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        credential_id: string, optional
            The ID of the set of credentials to
            use instead of user and password. Note that with this change, username and password
            will become optional.
        use_kerberos: Optional[bool]
            If unset, uses the server default: False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        base_data = {
            "data_source_id": data_source_id,
            "user": username,
            "password": password,
            "categories": categories,
            "credential_id": credential_id,
            "use_kerberos": use_kerberos,
            "credential_data": credential_data,
        }
        data = _remove_empty_params(
            params_dict=base_data,
            required_params={"data_source_id"},
        )

        if "credential_data" in data:
            data["credential_data"] = CredentialDataSchema(data["credential_data"])

        upload_url = f"{cls._path}{dataset_id}/versions/fromDataSource/"
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)


def _remove_empty_params(  # pylint: disable=missing-function-docstring
    params_dict: Dict[str, Any],
    required_params: Optional[Set[str]] = None,
) -> Dict[str, Any]:
    non_empty_params = {key: value for key, value in params_dict.items() if value is not None}
    if required_params and not required_params.issubset(set(non_empty_params.keys())):
        required_but_none = sorted(required_params.difference(set(non_empty_params.keys())))
        raise InvalidUsageError(
            f"Missing required parameters ({', '.join(required_but_none)}) must be set."
        )
    return non_empty_params


def _update_filter_failed(query_params: Dict[str, str]) -> None:
    try:
        key = "filter_failed"
        query_params[key] = str(query_params[key]).lower()
    except KeyError:
        pass


def _safe_merge(first: t.Dict, second: t.Dict) -> t.Dict:
    second_names = {el.name for el in second.keys}
    if any(el.name in second_names for el in first.keys):
        raise ValueError("Duplicate keys detected")

    return first.merge(second)


class DatasetDetails(APIObject):
    """Represents a detailed view of a Dataset. The `to_dataset` method creates a Dataset
    from this details view.

    Attributes
    ----------
    dataset_id: string
        The ID of this dataset
    name: string
        The name of this dataset in the catalog
    is_latest_version: bool
        Whether this dataset version is the latest version
        of this dataset
    version_id: string
        The object ID of the catalog_version the dataset belongs to
    categories: list(string)
        An array of strings describing the intended use of the dataset. The
        supported options are "TRAINING" and "PREDICTION".
    created_at: string
        The date when the dataset was created
    created_by: string
        Username of the user who created the dataset
    is_snapshot: bool
        Whether the dataset version is an immutable snapshot of data
        which has previously been retrieved and saved to Data_robot
    data_persisted: Optional[bool]
        If true, user is allowed to view extended data profile
        (which includes data statistics like min/max/median/mean, histogram, etc.) and download
        data. If false, download is not allowed and only the data schema (feature names and types)
        will be available.
    is_data_engine_eligible: bool
        Whether this dataset can be
        a data source of a data engine query.
    processing_state: string
        Current ingestion process state of
        the dataset
    row_count: Optional[int]
        The number of rows in the dataset.
    size: Optional[int]
        The size of the dataset as a CSV in bytes.
    data_engine_query_id: string, optional
        ID of the source data engine query
    data_source_id: string, optional
        ID of the datasource used as the source of the dataset
    data_source_type: string
        the type of the datasource that was used as the source of the
        dataset
    description: string, optional
        the description of the dataset
    eda1_modification_date: string, optional
        the ISO 8601 formatted date and time when the EDA1 for
        the dataset was updated
    eda1_modifier_full_name: string, optional
        the user who was the last to update EDA1 for the
        dataset
    error: string
        details of exception raised during ingestion process, if any
    feature_count: Optional[int]
        total number of features in the dataset
    feature_count_by_type: list[FeatureTypeCount]
        number of features in the dataset grouped by feature type
    last_modification_date: string
        the ISO 8601 formatted date and time when the dataset
        was last modified
    last_modifier_full_name: string
        full name of user who was the last to modify the
        dataset
    tags: list[string]
        list of tags attached to the item
    uri: string
        the uri to datasource like:
        - 'file_name.csv'
        - 'jdbc:DATA_SOURCE_GIVEN_NAME/SCHEMA.TABLE_NAME'
        - 'jdbc:DATA_SOURCE_GIVEN_NAME/<query>' - for `query` based datasources
        - 'https://s3.amazonaws.com/my_data/my_dataset.csv'
        - etc.
    sample_size: dict, optional
        The size of data fetched during dataset registration. For example, to fetch the first 95
        rows,  the sample_size value is {'type': 'rows', 'value': 95}.
        Currently only 'rows' type is supported.
    """

    _extra_fields = t.Dict(
        {
            t.Key("data_engine_query_id", optional=True): String,
            t.Key("data_source_id", optional=True): String,
            t.Key("data_source_type"): String(allow_blank=True),
            t.Key("description", optional=True): String(allow_blank=True),
            t.Key("eda1_modification_date", optional=True): t.Call(dateutil.parser.parse),
            t.Key("eda1_modifier_full_name", optional=True): String,
            t.Key("error"): String(allow_blank=True),
            t.Key("feature_count", optional=True): Int,
            t.Key("feature_count_by_type", optional=True): t.List(
                t.Call(lambda d: FeatureTypeCount(**d))
            ),
            t.Key("last_modification_date"): t.Call(dateutil.parser.parse),
            t.Key("last_modifier_full_name"): String,
            t.Key("tags", optional=True): t.List(String),
            t.Key("uri"): String,
            t.Key("recipe_id", optional=True): String,
            t.Key("is_wrangling_eligible", optional=True): bool,
        }
    )

    _converter = _safe_merge(_extra_fields, _base_dataset_schema).allow_extra("*")

    _path = "datasets/"

    def __init__(
        self,
        dataset_id: str,
        version_id: str,
        categories: List[str],
        created_by: str,
        created_at: str,
        data_source_type: str,
        error: str,
        is_latest_version: bool,
        is_snapshot: bool,
        is_data_engine_eligible: bool,
        last_modification_date: str,
        last_modifier_full_name: str,
        name: str,
        uri: str,
        processing_state: str,
        data_persisted: Optional[bool] = None,
        data_engine_query_id: Optional[str] = None,
        data_source_id: Optional[str] = None,
        description: Optional[str] = None,
        eda1_modification_date: Optional[str] = None,
        eda1_modifier_full_name: Optional[str] = None,
        feature_count: Optional[int] = None,
        feature_count_by_type: Optional[List[FeatureTypeCount]] = None,
        row_count: Optional[int] = None,
        size: Optional[int] = None,
        tags: Optional[List[str]] = None,
        recipe_id: Optional[str] = None,
        is_wrangling_eligible: Optional[bool] = None,
        sample_size: Optional[Dict[str, Any]] = None,
    ):
        self.dataset_id = dataset_id
        self.version_id = version_id
        self.categories = categories
        self.created_by = created_by
        self.created_at = created_at
        self.data_source_type = data_source_type
        self.error = error
        self.is_latest_version = is_latest_version
        self.is_snapshot = is_snapshot
        self.is_data_engine_eligible = is_data_engine_eligible
        self.last_modification_date = last_modification_date
        self.last_modifier_full_name = last_modifier_full_name
        self.name = name
        self.uri = uri
        self.data_persisted = data_persisted
        self.data_engine_query_id = data_engine_query_id
        self.data_source_id = data_source_id
        self.description = description
        self.eda1_modification_date = eda1_modification_date
        self.eda1_modifier_full_name = eda1_modifier_full_name
        self.feature_count = feature_count
        self.feature_count_by_type = feature_count_by_type
        self.processing_state = processing_state
        self.row_count = row_count
        self.size = size
        self.tags = tags
        self.recipe_id = recipe_id
        self.is_wrangling_eligible = is_wrangling_eligible
        self.sample_size = sample_size

    @classmethod
    def get(cls: Type[TDatasetDetails], dataset_id: str) -> TDatasetDetails:
        """
        Get details for a Dataset from the server

        Parameters
        ----------
        dataset_id: str
            The id for the Dataset from which to get details

        Returns
        -------
        DatasetDetails
        """
        path = f"{cls._path}{dataset_id}/"
        return cls.from_location(path)

    def to_dataset(self) -> Dataset:
        """
        Build a Dataset object from the information in this object

        Returns
        -------
        Dataset
        """
        return Dataset(
            dataset_id=self.dataset_id,
            name=self.name,
            created_at=self.created_at,
            created_by=self.created_by,
            version_id=self.version_id,
            categories=self.categories,
            is_latest_version=self.is_latest_version,
            is_data_engine_eligible=self.is_data_engine_eligible,
            is_snapshot=self.is_snapshot,
            data_persisted=self.data_persisted,
            size=self.size,
            row_count=self.row_count,
            processing_state=self.processing_state,
        )


class DataStage(APIObject):
    """Represents a DataStage returned from the api/v2/dataStages/ endpoints.

    To create a data stage the user will need the Data Stages feature flag enabled.

    Attributes
    ----------
    id: string
        The ID of this dataset
    """

    _converter = t.Dict(
        {
            t.Key("id"): String,
        }
    ).allow_extra("*")

    url = "dataStages/"

    def __init__(
        self,
        id: str,
    ) -> None:
        self.id = id

    @classmethod
    def create_datastage(cls, filename: str) -> DataStage:
        """Create a data stage.

        To create a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        filename : str
            Name of the data stage
        """
        payload = {
            "filename": filename,
        }
        response = cls._client.post(cls.url, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def create_datastage_from_source(
        cls,
        filename: str,
        source: Union[str, pd.DataFrame, IOBase],
        max_chunk_size: Optional[int] = 100 * MEGABYTE,
    ) -> DataStage:
        """Create and finalize a new simple data stage from a DataFrame.

        To create a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        filename : str
            Name of the data stage
        source: str, pd.DataFrame, IOBase
            Pass a filepath, file-like object or DataFrame to create a data stage.
        max_chunk_size : Optional[int]
            Size of each chunk. Default is 100MB chunks.
        """
        source_type = parse_source_type(source)
        if source_type not in (
            FileLocationType.PATH,
            FileLocationType.URL,
            LocalSourceType.DATA_FRAME,
            LocalSourceType.FILELIKE,
        ):
            raise ValueError("Source type does not support automated data stage creation.")

        new_datastage = cls.create_datastage(filename=filename)

        if source_type == FileLocationType.PATH:
            new_datastage.upload_from_file(
                source=cast(str, source),
                max_chunk_size=max_chunk_size,
            )
        elif source_type == FileLocationType.URL:
            new_datastage.upload_from_url(source=cast(str, source), max_chunk_size=max_chunk_size)
        elif source_type == LocalSourceType.DATA_FRAME:
            new_datastage.upload_from_dataframe(source=source, max_chunk_size=max_chunk_size)
        elif source_type == LocalSourceType.FILELIKE:
            new_datastage._upload_contents_as_parts(
                contents=cast(IOBase, source), max_chunk_size=max_chunk_size
            )

        new_datastage.finalize()

        return new_datastage

    def create_dataset_from_datastage(
        self,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> Dataset:
        """
        A blocking call that creates a new Dataset from data stored as a DataStage.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which dataset creation is considered unsuccessful.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        return Dataset.create_from_datastage(
            datastage_id=self.id, categories=categories, max_wait=max_wait
        )

    def create_dataset_version_from_datastage(
        self,
        dataset_id: str,
        categories: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> Dataset:
        """
        A blocking call that creates a new Dataset from data stored as a DataStage for a given dataset.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: Optional[int]
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        return Dataset.create_version_from_datastage(
            dataset_id=dataset_id, datastage_id=self.id, categories=categories, max_wait=max_wait
        )

    def _upload_contents_as_parts(
        self,
        contents: Union[BinaryIO, IOBase],
        max_chunk_size: Optional[int] = 100 * MEGABYTE,
    ) -> None:
        """
        Upload a file-like object in chunks to a datastage.

        To upload to a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        contents : BinaryIO, IOBase
            File-like, readable object
        max_chunk_size : Optional[int]
            Size of each chunk. Default is 100MB chunks.
        """
        max_chunk_size = cast(int, max_chunk_size)

        part_number = 1
        while True:
            dat = contents.read(max_chunk_size)
            if not dat:
                break
            part_url = f"dataStages/{self.id}/parts/{part_number}/"
            encoder = MultipartEncoder(fields={"file": ("part", io.BytesIO(dat))})
            response = self._client.put(
                part_url,
                data=encoder,
                headers={"Content-Type": encoder.content_type},
            )
            assert response.status_code == 200
            part_number += 1

    def upload_from_dataframe(
        self, source: pd.DataFrame, max_chunk_size: Optional[int] = 100 * MEGABYTE
    ) -> None:
        """
        Upload all parts of a dataframe to a datastage. Parts will be automatically defined.

        To upload to a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        source : pd.DataFrame
            Pandas dataframe containing data
        max_chunk_size : Optional[int]
            Size of each chunk. Default is 100MB chunks.
        """
        source_type = parse_source_type(source)
        if source_type != LocalSourceType.DATA_FRAME:
            raise InvalidUsageError(
                "This upload function supports dataframes. See create_datastage_from_source."
            )

        contents = io.BytesIO()
        source.to_csv(contents)
        contents.seek(0)
        self._upload_contents_as_parts(contents, max_chunk_size=max_chunk_size)

    def upload_from_file(
        self,
        source: str,
        max_chunk_size: Optional[int] = 100 * MEGABYTE,
    ) -> None:
        """
        Upload all parts of a local file to a datastage. Parts will be automatically defined.

        To upload to a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        source : str
            Path to local file
        max_chunk_size : Optional[int]
            Size of each chunk. Default is 100MB chunks.
        """
        source_type = parse_source_type(source)
        if source_type != FileLocationType.PATH:
            raise InvalidUsageError(
                "This upload function supports file paths. See create_datastage_from_source."
            )

        with open(source, "rb") as contents:
            self._upload_contents_as_parts(contents, max_chunk_size=max_chunk_size)

    def upload_from_url(
        self,
        source: str,
        max_chunk_size: Optional[int] = 100 * MEGABYTE,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        stream_only: Optional[bool] = False,
    ) -> None:
        """
        This will download a file from a URL and upload it as a datastage. Parts will be automatically defined.
        Datastages do not support directly importing from a URL at this time, but this will help to make this
        process more simple.

        To upload to a data stage the user will need the Data Stages feature flag enabled.

        Parameters
        ----------
        source : str
            URL of the file
        max_chunk_size : Optional[int]
            Size of each chunk. Default is 100MB chunks.
        max_wait : Optional[int]
            Time in seconds after which model import is considered unsuccessful
        stream_only : Optional[bool]
            Stream the file only without using temporary local storage. This method is not recommended for larger
            files. It will keep the entire request in ram during the transfer.
        """
        source_type = parse_source_type(source)
        if source_type != FileLocationType.URL:
            raise InvalidUsageError(
                "This upload function supports urls. See create_datastage_from_source."
            )

        if not stream_only:
            with tempfile.NamedTemporaryFile() as temporary_file:
                with requests.get(source, timeout=max_wait, stream=True) as req:
                    req.raise_for_status()
                    for chunk in req.iter_content(chunk_size=8192):
                        temporary_file.write(chunk)
                temporary_file.flush()
                temporary_file.seek(0)
                self._upload_contents_as_parts(
                    cast(IOBase, temporary_file), max_chunk_size=max_chunk_size
                )
        else:
            downloaded_file = requests.get(source, timeout=max_wait)
            self._upload_contents_as_parts(
                BytesIO(downloaded_file.content), max_chunk_size=max_chunk_size
            )

    def finalize(self) -> Dict[str, List[Dict[str, str]]]:
        """
        Finalize a data stage.

        To finalize a data stage the user will need the Data Stages feature flag enabled.

        Returns
        -------
        response: Response
            The Dataset version created from the uploaded data
        """
        url = f"{self.url}{self.id}/finalize/"
        response = self._client.post(url)
        assert response.status_code == 200

        parts = list(response.json()["parts"])
        return {"parts": parts}



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot.enums import enum_to_list, FeedbackSentiment
from datarobot.models.api_object import APIObject
from datarobot.models.genai.chat import Chat
from datarobot.models.genai.llm import LLMDefinition
from datarobot.models.genai.llm_blueprint import (
    LLMBlueprint,
    vector_database_settings_trafaret,
    VectorDatabaseSettings,
)
from datarobot.models.genai.playground import Playground
from datarobot.models.genai.vector_database import VectorDatabase
from datarobot.utils import rawdict
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def get_entity_id(
    entity: Union[Playground, Chat, ChatPrompt, LLMBlueprint, LLMDefinition, VectorDatabase, str]
) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


confidence_scores_trafaret = t.Dict(
    {
        t.Key("rouge"): t.Float,
        t.Key("meteor"): t.Float,
        t.Key("bleu"): t.Float,
    }
).ignore_extra("*")


metric_metadata_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("value"): t.Any,
        t.Key("formatted_value", optional=True): t.Or(t.String, t.Null),
        t.Key("sidecar_model_metric_validation_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_id", optional=True): t.Or(t.String, t.Null),
        t.Key("evaluation_dataset_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("cost_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_is_deprecated", optional=True): t.Or(t.Bool, t.Null),
    }
).ignore_extra("*")


feedback_result_trafaret = t.Dict(
    {
        t.Key("positive_user_ids"): t.List(t.String),
        t.Key("negative_user_ids"): t.List(t.String),
    }
).ignore_extra("*")

citation_trafaret = t.Dict(
    {
        t.Key("text"): t.String,
        t.Key("source", optional=True): t.Or(t.String, t.Null),
        t.Key("similarity_score", optional=True): t.Or(t.Float, t.Null),
        t.Key("metadata", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
    }
).ignore_extra("*")


result_metadata_trafaret = t.Dict(
    {
        t.Key("cost", optional=True): t.Or(t.Float, t.Null),
        t.Key("output_token_count"): t.Int,
        t.Key("input_token_count"): t.Int,
        t.Key("total_token_count"): t.Int,
        t.Key("estimated_docs_token_count"): t.Int,
        t.Key("latency_milliseconds"): t.Int,
        t.Key("error_message", optional=True): t.Or(t.String, t.Null),
        t.Key("feedback_result"): feedback_result_trafaret,
        t.Key("metrics"): t.List(metric_metadata_trafaret),
        t.Key("final_prompt", optional=True): t.Or(
            t.String,
            t.Dict().allow_extra("*"),
            t.List(t.Dict().allow_extra("*")),
            t.Null,
        ),
    }
).ignore_extra("*")

feedback_metadata_trafaret = t.Dict(
    {t.Key("feedback"): t.Enum(*enum_to_list(FeedbackSentiment))}
).ignore_extra("*")

chat_prompt_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("text"): t.String(allow_blank=True),
        t.Key("llm_blueprint_id"): t.String,
        t.Key("llm_id"): t.String,
        t.Key("llm_settings", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("vector_database_id", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_settings", optional=True): t.Or(
            vector_database_settings_trafaret, t.Null
        ),
        t.Key("result_metadata", optional=True): t.Or(result_metadata_trafaret, t.Null),
        t.Key("result_text", optional=True): t.Or(t.String(allow_blank=True), t.Null),
        t.Key("confidence_scores", optional=True): t.Or(confidence_scores_trafaret, t.Null),
        t.Key("citations"): t.List(citation_trafaret),
        t.Key("execution_status"): t.String,
        t.Key("chat_id", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_context_id", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_prompt_ids_included_in_history", optional=True): t.Or(t.List(t.String), t.Null),
        t.Key("metadata_filter", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
    }
).ignore_extra("*")


class FeedbackMetadataDict(TypedDict):
    feedback: FeedbackSentiment


class MetricMetadataDict(TypedDict):
    name: str
    value: Any
    formatted_value: Optional[str]
    sidecar_model_metric_validation_id: Optional[str]
    custom_model_id: Optional[str]
    evaluation_dataset_configuration_id: Optional[str]
    cost_configuration_id: Optional[str]
    llm_is_deprecated: Optional[bool]


class ResultMetadata(APIObject):
    """
    Metadata for the result of a chat prompt submission.

    Attributes
    ----------
    output_token_count : int
        The number of tokens in the output.
    input_token_count : int
        The number of tokens in the input. This includes the chat history and documents
        retrieved from a vector database, if any.
    total_token_count : int
        The total number of tokens processed.
    estimated_docs_token_count : int
        The estimated number of tokens from the documents retrieved from a vector database, if any.
    latency_milliseconds : int
        The latency of the chat prompt submission in milliseconds.
    feedback_result : FeedbackResult
        The lists of user_ids providing positive and negative feedback.
    metrics : MetricMetadata
        The evaluation metrics for this prompt.
    final_prompt : Optional[Union[str, dict]], optional
        Representation of the final prompt sent to the LLM.
    error_message : str or None, optional
        The error message from the LLM response.
    cost : float or None, optional
        The cost of the chat prompt submission.
    """

    _converter = result_metadata_trafaret

    def __init__(
        self,
        output_token_count: int,
        input_token_count: int,
        total_token_count: int,
        estimated_docs_token_count: int,
        latency_milliseconds: int,
        feedback_result: Dict[str, Any],
        metrics: List[Dict[str, Any]],
        final_prompt: Optional[Union[str, Dict[str, Any], List[Dict[str, Any]]]] = None,
        error_message: Optional[str] = None,
        cost: Optional[float] = None,
    ):
        self.output_token_count = output_token_count
        self.input_token_count = input_token_count
        self.total_token_count = total_token_count
        self.estimated_docs_token_count = estimated_docs_token_count
        self.latency_milliseconds = latency_milliseconds
        self.error_message = error_message
        self.cost = cost
        self.feedback_result = FeedbackResult.from_server_data(feedback_result)
        self.metrics = [
            MetricMetadata.from_server_data(metric_metadata) for metric_metadata in metrics
        ]
        self.final_prompt = final_prompt

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(output_token_count={self.output_token_count}, "
            f"input_token_count={self.input_token_count}, "
            f"total_token_count={self.total_token_count}, "
            f"estimated_docs_token_count={self.estimated_docs_token_count}, "
            f"latency_milliseconds={self.latency_milliseconds})"
        )


class ConfidenceScores(APIObject):
    """
    Confidence scores for a chat prompt submission that uses a vector database.

    Attributes
    ----------
    rouge : float
        The rouge score
    meteor : float
        The meteor score
    bleu : float
        The bleu score
    """

    _converter = confidence_scores_trafaret

    def __init__(
        self,
        rouge: float,
        meteor: float,
        bleu: float,
    ):
        self.rouge = rouge
        self.meteor = meteor
        self.bleu = bleu

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(rouge={self.rouge}, "
            f"meteor={self.meteor}, bleu={self.bleu})"
        )


class FeedbackMetadata(APIObject):
    """
    Metadata for feedback associated with a chat prompt.

    Attributes
    ----------
    feedback : FeedbackSentiment
        The sentiment of the feedback.
    """

    _converter = feedback_metadata_trafaret

    def __init__(self, feedback: FeedbackSentiment):
        self.feedback = feedback

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(feedback={self.feedback})"

    def to_dict(self) -> FeedbackMetadataDict:
        return {"feedback": self.feedback}


class FeedbackResult(APIObject):
    """
    Feedback associated with a chat prompt.

    Attributes
    ----------
    positive_user_ids : List[str]
        The IDs of the users providing positive feedback.
    negative_user_ids : List[str]
        The IDs of the users providing negative feedback.
    """

    _converter = feedback_result_trafaret

    def __init__(self, positive_user_ids: List[str], negative_user_ids: List[str]):
        self.positive_user_ids = positive_user_ids
        self.negative_user_ids = negative_user_ids

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class MetricMetadata(APIObject):
    """
    Metadata for the custom metrics associated with a chat prompt.

    Attributes
    ----------
    name : str
        The name of the metric.
    value : Any
        The value of the metric.
    formatted_value : Optional[str], optional
        The formatted value of the metric, if any.
    sidecar_model_metric_validation_id : Optional[str], optional
        The ID of the sidecar model implementing the metric, if any.
    custom_model_id : Optional[str], optional
        The ID of the custom model implementing the metric, if any.
    evaluation_dataset_configuration_id : Optional[str], optional
        The ID of the evaluation dataset configuration associated with the metric, if any.
    cost_configuration_id : Optional[str], optional
        The ID of the cost configuration associated with the metric, if any.
    llm_is_deprecated : Optional[bool], optional
        Whether the LLM is deprecated and will be removed in a future release.

    """

    _converter = metric_metadata_trafaret

    def __init__(
        self,
        name: str,
        value: Any,
        formatted_value: Optional[str] = None,
        sidecar_model_metric_validation_id: Optional[str] = None,
        custom_model_id: Optional[str] = None,
        evaluation_dataset_configuration_id: Optional[str] = None,
        cost_configuration_id: Optional[str] = None,
        llm_is_deprecated: Optional[bool] = None,
    ):
        self.name = name
        self.value = value
        self.formatted_value = formatted_value
        self.sidecar_model_metric_validation_id = sidecar_model_metric_validation_id
        self.custom_model_id = custom_model_id
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.cost_configuration_id = cost_configuration_id
        self.llm_is_deprecated = llm_is_deprecated

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name}, "
            f"value={self.value}, "
            f"formatted_value={self.formatted_value})"
        )

    def to_dict(self) -> MetricMetadataDict:
        return {
            "name": self.name,
            "value": self.value,
            "formatted_value": self.formatted_value,
            "sidecar_model_metric_validation_id": self.sidecar_model_metric_validation_id,
            "custom_model_id": self.custom_model_id,
            "evaluation_dataset_configuration_id": self.evaluation_dataset_configuration_id,
            "cost_configuration_id": self.cost_configuration_id,
            "llm_is_deprecated": self.llm_is_deprecated,
        }


class Citation(APIObject):
    """
    Citation for documents retrieved from a vector database.

    Attributes
    ----------
    text : str
        The text retrieved from a vector database.
    source : str or None, optional
        The source of the retrieved text.
    similarity_score:
        The similarity score between the citation and the user prompt.
    metadata: dict or None, optional
        Additional metadata for the citation.
    """

    _converter = citation_trafaret

    def __init__(
        self,
        text: str,
        source: Optional[str] = None,
        similarity_score: Optional[float] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.text = text
        self.source = source
        self.similarity_score = similarity_score
        self.metadata = metadata

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={self.text}, source={self.source})"


class ChatPrompt(APIObject):
    """
    Metadata for a DataRobot GenAI chat prompt.

    Attributes
    ----------
    id : str
        Chat prompt ID.
    text : str
        The prompt text.
    llm_blueprint_id : str
        ID of the LLM blueprint associated with the chat prompt.
    llm_id : str
        ID of the LLM type. This must be one of the IDs returned by `LLMDefinition.list`
        for this user.
    llm_settings : dict or None
        The LLM settings for the LLM blueprint. The specific keys allowed and the
        constraints on the values are defined in the response from `LLMDefinition.list`,
        but this typically has dict fields. Either:
        - system_prompt - The system prompt that influences the LLM responses.
        - max_completion_length - The maximum number of tokens in the completion.
        - temperature - Controls the variability in the LLM response.
        - top_p - Sets whether the model considers next tokens with top_p probability mass.
        or
        - system_prompt - The system prompt that influences the LLM responses.
        - validation_id - The ID of the external model LLM validation.
        - external_llm_context_size - The external LLM's context size, in tokens,
        for external model-based LLM blueprints.
    creation_date : str
        The date the chat prompt was created.
    creation_user_id : str
        ID of the creating user.
    vector_database_id : str or None
        ID of the vector database associated with the LLM blueprint, if any.
    vector_database_settings : VectorDatabaseSettings or None
        The settings for the vector database associated with the LLM blueprint, if any.
    result_metadata : ResultMetadata or None
        Metadata for the result of the chat prompt submission.
    result_text: str or None
        The result text from the chat prompt submission.
    confidence_scores: ConfidenceScores or None
        The confidence scores if there is a vector database associated with the chat prompt.
    citations: list[Citation]
        List of citations from text retrieved from the vector database, if any.
    execution_status: str
        The execution status of the chat prompt.
    chat_id: Optional[str]
        ID of the chat associated with the chat prompt.
    chat_context_id: Optional[str]
        The ID of the chat context for the chat prompt.
    chat_prompt_ids_included_in_history: Optional[list[str]]
        The IDs of the chat prompts included in the chat history for this chat prompt.
    metadata_filter: Optional[Dict[str, Any] | None]
        The metadata filter to apply to the vector database.

        Supports:
        - None or empty dict (no filters): Considers all documents
        - Multiple field filters (implicit AND): {"a": 1, "b": "b"}
        - Comparison operators: {"field": {"$gt": 5}}
        - Logical operators: {"$and": [...], "$or": [...]}
        - Nested combinations of the above

        Comparison operators:
        - $eq: equal to (string, int, float, bool)
        - $ne: not equal to (string, int, float, bool)
        - $gt: greater than (int, float)
        - $gte: greater than or equal to (int, float)
        - $lt: less than (int, float)
        - $lte: less than or equal to (int, float)
        - $in: a value is in list (string, int, float, bool)
        - $nin: a value is not in list (string, int, float, bool)
        - $contains: a string contains a value (string)
        - $not_contains: a string does not contain a value (string)
    """

    _path = "api/v2/genai/chatPrompts"

    _converter = chat_prompt_trafaret

    def __init__(
        self,
        id: str,
        text: str,
        llm_blueprint_id: str,
        llm_id: str,
        creation_date: str,
        creation_user_id: str,
        citations: List[Dict[str, Any]],
        execution_status: str,
        llm_settings: Optional[Dict[str, Any]] = None,
        vector_database_id: Optional[str] = None,
        vector_database_settings: Optional[Dict[str, Any]] = None,
        result_metadata: Optional[Dict[str, Any]] = None,
        result_text: Optional[str] = None,
        confidence_scores: Optional[Dict[str, float]] = None,
        chat_id: Optional[str] = None,
        chat_context_id: Optional[str] = None,
        chat_prompt_ids_included_in_history: Optional[list[str]] = None,
        metadata_filter: Optional[Dict[str, Any]] = None,
    ):
        self.id = id
        self.text = text
        self.llm_blueprint_id = llm_blueprint_id
        self.llm_id = llm_id
        self.llm_settings = llm_settings
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.citations = [Citation.from_server_data(citation) for citation in citations]
        self.execution_status = execution_status
        self.vector_database_id = vector_database_id
        self.vector_database_settings = (
            VectorDatabaseSettings.from_server_data(vector_database_settings)
            if vector_database_settings
            else None
        )
        self.result_metadata = (
            ResultMetadata.from_server_data(result_metadata) if result_metadata else None
        )
        self.result_text = result_text
        self.confidence_scores = (
            ConfidenceScores.from_server_data(confidence_scores) if confidence_scores else None
        )
        self.chat_id = chat_id
        self.chat_context_id = chat_context_id
        self.chat_prompt_ids_included_in_history = chat_prompt_ids_included_in_history
        self.citations = [Citation.from_server_data(citation) for citation in citations]
        self.vector_database_settings = (
            VectorDatabaseSettings.from_server_data(vector_database_settings)
            if vector_database_settings
            else None
        )
        self.result_metadata = (
            ResultMetadata.from_server_data(result_metadata) if result_metadata else None
        )
        self.metadata_filter = metadata_filter

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, "
            f"execution_status={self.execution_status}, text={self.text[:1000]})"
        )

    @classmethod
    def create(
        cls,
        text: str,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
        chat: Optional[Union[Chat, str]] = None,
        llm: Optional[Union[LLMDefinition, str]] = None,
        llm_settings: Optional[Dict[str, Optional[Union[bool, int, float, str]]]] = None,
        vector_database: Optional[Union[VectorDatabase, str]] = None,
        vector_database_settings: Optional[VectorDatabaseSettings] = None,
        wait_for_completion: bool = False,
        metadata_filter: Optional[Dict[str, Any]] = None,
    ) -> ChatPrompt:
        """
        Create a new ChatPrompt. This submits the prompt text to the LLM. Either `llm_blueprint`
        or `chat` is required.

        Parameters
        ----------
        text : str
            The prompt text.
        llm_blueprint : LLMBlueprint or str or None, optional
            The LLM blueprint associated with the created chat prompt, either `LLMBlueprint` or
            LLM blueprint ID.
        chat : Chat or str or None, optional
            The chat associated with the created chat prompt, either `Chat` or chat ID.
        llm : LLMDefinition, str, or None, optional
            The LLM to use for the chat prompt, either `LLMDefinition` or LLM blueprint ID.
        llm_settings: dict or None
            LLM settings to use for the chat prompt. The specific keys allowed and the
            constraints on the values are defined in the response from `LLMDefinition.list`
            but this typically has dict fields:
            - system_prompt - The system prompt that tells the LLM how to behave.
            - max_completion_length - The maximum number of tokens in the completion.
            - temperature - Controls the variability in the LLM response.
            - top_p - Whether the model considers next tokens with top_p probability mass.
            Or
            - system_prompt - The system prompt that tells the LLM how to behave.
            - validation_id - The ID of the custom model LLM validation
            for custom model LLM blueprints.
        vector_database: VectorDatabase, str, or None, optional
            The vector database to use with this chat prompt submission, either
            `VectorDatabase` or vector database ID.
        vector_database_settings: VectorDatabaseSettings or None, optional
            Settings for the vector database, if any.
        wait_for_completion : bool
            If set to True, chat prompt result response limit is up to 10 minutes,
            raising a timeout error after that.
            Otherwise, check current status by using `ChatPrompt.get` with returned ID.
        metadata_filter: Optional[Dict[str, Any] | None]
            The metadata filter to apply to the vector database.

            Supports:
            - None or empty dict (no filters): Considers all documents
            - Multiple field filters (implicit AND): {"a": 1, "b": "b"}
            - Comparison operators: {"field": {"$gt": 5}}
            - Logical operators: {"$and": [...], "$or": [...]}
            - Nested combinations of the above

            Comparison operators:
            - $eq: equal to (string, int, float, bool)
            - $ne: not equal to (string, int, float, bool)
            - $gt: greater than (int, float)
            - $gte: greater than or equal to (int, float)
            - $lt: less than (int, float)
            - $lte: less than or equal to (int, float)
            - $in: a value is in list (string, int, float, bool)
            - $nin: a value is not in list (string, int, float, bool)
            - $contains: a string contains a value (string)
            - $not_contains: a string does not contain a value (string)

        Returns
        -------
        chat_prompt : ChatPrompt
            The created chat prompt.
        """
        payload = {
            "llm_blueprint_id": get_entity_id(llm_blueprint) if llm_blueprint else None,
            "chat_id": get_entity_id(chat) if chat else None,
            "text": text,
            "llm_id": get_entity_id(llm) if llm else None,
            "llm_settings": llm_settings,
            "vector_database_id": get_entity_id(vector_database) if vector_database else None,
            "vector_database_settings": (
                vector_database_settings.to_dict() if vector_database_settings else None
            ),
            "metadata_filter": rawdict(metadata_filter) if metadata_filter else None,
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, r_data.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(r_data.json())

    def update(
        self,
        custom_metrics: Optional[list[MetricMetadata]] = None,
        feedback_metadata: Optional[FeedbackMetadata] = None,
    ) -> ChatPrompt:
        """
        Update the chat prompt.

        Parameters
        ----------
        custom_metrics : Optional[list[MetricMetadata]], optional
            The new custom metrics to add to the chat prompt.
        feedback_metadata: Optional[FeedbackMetadata], optional
            The new feedback to add to the chat prompt.

        Returns
        -------
        chat_prompt : ChatPrompt
            The updated chat prompt.
        """
        payload = {
            "custom_metrics": (
                [metadata.to_dict() for metadata in custom_metrics] if custom_metrics else None
            ),
            "feedback_metadata": feedback_metadata.to_dict() if feedback_metadata else None,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())

    @classmethod
    def get(cls, chat_prompt: Union[ChatPrompt, str]) -> ChatPrompt:
        """
        Retrieve a single chat prompt.

        Parameters
        ----------
        chat_prompt : ChatPrompt or str
            The chat prompt you want to retrieve, either `ChatPrompt` or chat prompt ID.

        Returns
        -------
        chat_prompt : ChatPrompt
            The requested chat prompt.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(chat_prompt)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
        playground: Optional[Union[Playground, str]] = None,
        chat: Optional[Union[Chat, str]] = None,
    ) -> List[ChatPrompt]:
        """
        List all chat prompts available to the user. If the `llm_blueprint`, `playground`, or `chat`
        is specified then the results are restricted to the chat prompts associated with that
        entity.

        Parameters
        ----------
        llm_blueprint : Optional[Union[LLMBlueprint, str]], optional
            The returned chat prompts are filtered to those associated with a specific LLM blueprint
            if it is specified. Accepts either `LLMBlueprint` or LLM blueprint ID.
        playground : Optional[Union[Playground, str]], optional
            The returned chat prompts are filtered to those associated with a specific playground
            if it is specified. Accepts either `Playground` or playground ID.
        chat : Optional[Union[Chat, str]], optional
            The returned chat prompts are filtered to those associated with a specific chat
            if it is specified. Accepts either `Chat` or chat ID.

        Returns
        -------
        chat_prompts : list[ChatPrompt]
            A list of chat prompts available to the user.
        """
        params = {
            "llm_blueprint_id": get_entity_id(llm_blueprint) if llm_blueprint else None,
            "playground_id": get_entity_id(playground) if playground else None,
            "chat_id": get_entity_id(chat) if chat else None,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete the single chat prompt.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    def create_llm_blueprint(self, name: str, description: str = "") -> LLMBlueprint:
        """
        Create a new LLM blueprint from an existing chat prompt.

        Parameters
        ----------
        name : str
            LLM blueprint name.
        description : Optional[str]
            Description of the LLM blueprint, by default "".

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The created LLM blueprint.
        """
        payload = {
            "chat_prompt_id": self.id,
            "name": name,
            "description": description,
        }

        url = f"{self._client.domain}/{LLMBlueprint._path}/fromChatPrompt/"
        r_data = self._client.post(url, data=payload)
        return LLMBlueprint.from_server_data(r_data.json())



# -*- coding: utf-8 -*-
#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc. Confidential.
#
# This is unpublished proprietary source code of DataRobot, Inc.
# and its affiliates.
#
# The copyright notice above does not evidence any actual or intended
# publication of such source code.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

import trafaret as t

from datarobot.models.api_object import APIObject

T = TypeVar("T", bound="FairnessInsights")


@dataclass
class PerClassFairnessData:
    class_name: str
    value: float
    absolute_value: float
    entries_count: int
    is_statistically_significant: bool


@dataclass
class FairnessInsightData:
    fairness_metric: str
    fairness_threshold: str
    prediction_threshold: float
    protected_feature: str
    per_class_fairness: List[PerClassFairnessData]


class FairnessInsights(APIObject):
    """
    Bias & Fairness - Per-Class Fairness Insight computed data.

    Attributes
    ----------
    model_id: str
        The model ID for which the Fairness Insight is computed.

    count: int
        The number of per-class fairness insights data computed for the given model.

    insights: List[FairnessInsightData]
        A list of computed per-class fairness insights data.
    """

    _path = "/projects/{}/models/{}/fairnessInsights/"

    per_class_fairness_trafaret = t.Dict(
        {
            t.Key("class_name"): t.String(),
            t.Key("value"): t.Float(),
            t.Key("absolute_value"): t.Float(),
            t.Key("entries_count"): t.Int,
            t.Key("is_statistically_significant"): t.Bool,
        }
    )

    single_fairness_insight_trafaret = t.Dict(
        {
            t.Key("fairness_metric"): t.Or(t.String(), t.Null),
            t.Key("fairness_threshold", optional=True, default=0.8): t.Float(),
            t.Key("prediction_threshold"): t.Or(t.Float(), t.Null),
            t.Key("protected_feature"): t.Or(t.String(), t.Null),
            t.Key("per_class_fairness"): t.List(per_class_fairness_trafaret),
        }
    )

    _converter = t.Dict(
        {
            t.Key("data"): t.List(
                t.Dict({t.Key("model_id"): t.String()}).merge(single_fairness_insight_trafaret)
            )
        }
    )

    def __init__(self, data: List[Dict[str, Any]]) -> None:
        self.model_id = [insight_data.pop("model_id") for insight_data in data][0]
        self.count = len(data)
        self.insights = [
            FairnessInsightData(**insight_data)
            for insight_data in data
            if self.single_fairness_insight_trafaret.check(insight_data)
        ]

    @classmethod
    def from_server_data(  # type: ignore[override]
        cls: Type[T],
        data: Dict[str, List[Dict[str, Any]]],
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> FairnessInsights:
        if data.get("count"):
            data = {"data": data["data"]}
        return super().from_server_data(data=data, keep_attrs=keep_attrs)

    def __repr__(self) -> str:
        return f"FairnessInsights({self.model_id})"



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datarobot.models.genai.custom_model_validation import NonChatAwareCustomModelValidation


class CustomModelEmbeddingValidation(NonChatAwareCustomModelValidation):
    """
    Validation record checking the ability of the deployment to serve as a custom model embedding.

    Attributes
    ----------
    id : str
        The ID of the validation.
    prompt_column_name : str
        The column name the deployed model expect as the input.
    target_column_name : str
        The target name that the deployed model will output.
    deployment_id : str
        The ID of the deployment.
    model_id : str
        The ID of the underlying deployed model, which can be found using `Deployment.model["id"]`.
    validation_status : str
        Can be TESTING, FAILED, or PASSED. Only PASSED is allowed for use.
    deployment_access_data : dict, optional
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The creating user's tenant ID.
    name : str
        The display name of the validated custom model.
    creation_date : str
        The creation date of the validation (ISO 8601 formatted).
    user_id : str
        The ID of the creating user.
    error_message : Optional[str]
        Additional information for the errored validation.
    deployment_name : Optional[str]
        The name of the validated deployment.
    user_name : Optional[str]
        The name of the creating user.
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    prediction_timeout: int
        The timeout in seconds for the prediction API used in this custom model validation.
    """

    _path = "api/v2/genai/customModelEmbeddingValidations"



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.llm_blueprint import LLMSettingsCommonDict, LLMSettingsCustomModelDict

synthetic_dataset_generation_response_trafaret = t.Dict(
    {
        t.Key("dataset_id"): t.String,
        t.Key("prompt_column_name"): t.String,
        t.Key("response_column_name"): t.String,
    }
).ignore_extra("*")


class SyntheticEvaluationDataset(APIObject):
    """A synthetically generated evaluation dataset for LLMs.

    Attributes
    ----------
    dataset_id (str): The ID of the dataset.
    prompt_column_name (str): The name of the prompt column in the dataset.
    response_column_name (str): The name of the response column in the dataset.
    """

    _path = "api/v2/genai/syntheticEvaluationDatasetGenerations/"

    _converter = synthetic_dataset_generation_response_trafaret

    def __init__(self, dataset_id: str, prompt_column_name: str, response_column_name: str) -> None:
        self.dataset_id = dataset_id
        self.prompt_column_name = prompt_column_name
        self.response_column_name = response_column_name

    @classmethod
    def create(
        cls,
        llm_id: str,
        vector_database_id: str,
        llm_settings: Optional[Union[LLMSettingsCommonDict, LLMSettingsCustomModelDict]] = None,
        dataset_name: Optional[str] = None,
        language: Optional[str] = None,
    ) -> SyntheticEvaluationDataset:
        """Create a synthetic evaluation dataset generation job.  This will
        create a synthetic dataset to be used for evaluation of a language model.

        Parameters
        ----------
        llm_id (str): The ID of the language model that will be used to generate the dataset.
        llm_settings (Dict[Optional[str][Union[bool, int, float, str]]]): The settings to use for the language
                model used for dataset generation.
        vector_database_id (str): The ID of the vector database to use for dataset generation.
        dataset_name (str): The name to use for the dataset, otherwise this will be autogenerated.
        language (str): The language to use for dataset generation.  English will be used if this is not specified.

        Returns
        -------
        SyntheticEvaluationDataset: Reference to the synthetic evaluation dataset that was created.

        """
        url = f"{cls._client.domain}/{cls._path}"
        payload = {
            "llm_id": llm_id,
            "llm_settings": llm_settings,
            "vector_database_id": vector_database_id,
            "dataset_name": dataset_name,
            "language": language,
        }
        response_data = cls._client.post(url, payload)
        return cls.from_server_data(response_data.json())

    def __repr__(self) -> str:
        return (
            f"<SyntheticEvaluationDataset {self.dataset_id}: {self.prompt_column_name} "
            "-> {self.response_column_name}>"
        )



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Optional

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject


class CustomTaskVersionDependencyBuild(APIObject):
    """Metadata about a DataRobot custom task version's dependency build
    .. versionadded:: v2.27

    Attributes
    ----------
    custom_task_id: str
        id of the custom task
    custom_task_version_id: str
        id of the custom task version
    build_status: str
        the status of the custom task version's dependency build
    started_at: str
        ISO-8601 formatted timestamp of when the build was started
    completed_at: str
        ISO-8601 formatted timestamp of when the build has completed
    build_log_location: str
        Location of retrieving dependency build log
    """

    _converter = t.Dict(
        {
            t.Key("custom_task_id"): String(),
            t.Key("custom_task_version_id"): String(),
            t.Key("build_status"): String(),
            t.Key("build_start") >> "started_at": String(),
            t.Key("build_end", optional=True) >> "completed_at": t.Or(String(), t.Null()),
            t.Key("build_log_location", optional=True): t.Or(String(), t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        custom_task_id: str,
        custom_task_version_id: str,
        build_status: str,
        started_at: str,
        completed_at: Optional[str] = None,
        build_log_location: Optional[str] = None,
    ) -> None:
        self.custom_task_id = custom_task_id
        self.custom_task_version_id = custom_task_version_id
        self.build_status = build_status
        self.started_at = started_at
        self.completed_at = completed_at
        self.build_log_location = build_log_location

    def __repr__(self) -> str:
        return "{}(task={!r}, version={!r}, status={!r})".format(
            self.__class__.__name__,
            self.custom_task_id,
            self.custom_task_version_id,
            self.build_status,
        )



#
# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from io import IOBase
from pathlib import Path
from typing import Union
from urllib.parse import urlparse

from pandas import DataFrame

from datarobot.enums import FileLocationType, LocalSourceType
from datarobot.errors import InvalidUsageError


def parse_source_type(
    source: Union[str, DataFrame, IOBase],
) -> str:
    """Utility to check source type and return its type"""
    if isinstance(source, str):
        parse_result = urlparse(source)
        if parse_result.scheme and parse_result.netloc:
            return FileLocationType.URL
        elif Path(source).is_file():
            return FileLocationType.PATH
        else:
            raise InvalidUsageError(f"Unable to parse source ({source}) as URL or filepath.")
    elif isinstance(source, DataFrame):
        return LocalSourceType.DATA_FRAME
    elif isinstance(source, IOBase):
        return LocalSourceType.FILELIKE
    else:
        raise InvalidUsageError(
            f"Unable to parse source ({source}) as URL, filepath, DataFrame or file."
        )
